@@ -1,4 +1,7 @@
+from inspect import isclass
 from django.contrib.admin import register, ModelAdmin, site as admin_site
+from django.db.models.query import QuerySet
+from django.http import HttpRequest
 from polymorphic.admin import (
     StackedPolymorphicInline,
     PolymorphicInlineSupportMixin,
@@ -34,21 +37,29 @@ admin_site.register(Model2D, PolymorphicChildModelAdmin)
 class PlainAAdmin(ModelAdmin):
     search_fields = ["field1"]
 
-
-class InlineModelAChild(StackedPolymorphicInline.Child):
-    model = InlineModelA
-
-
-class InlineModelBChild(StackedPolymorphicInline.Child):
-    model = InlineModelB
-    autocomplete_fields = ["plain_a"]
+    def get_queryset(self, request: HttpRequest) -> QuerySet:
+        return super().get_queryset(request).order_by("pk")
 
 
 class Inline(StackedPolymorphicInline):
     model = InlineModelA
-    child_inlines = (InlineModelAChild, InlineModelBChild)
+
+    def get_child_inlines(self):
+        return [
+            child
+            for child in self.__class__.__dict__.values()
+            if isclass(child) and issubclass(child, StackedPolymorphicInline.Child)
+        ]
+
+    class InlineModelAChild(StackedPolymorphicInline.Child):
+        model = InlineModelA
+
+    class InlineModelBChild(StackedPolymorphicInline.Child):
+        model = InlineModelB
+        autocomplete_fields = ["plain_a"]
 
 
 @register(InlineParent)
 class InlineParentAdmin(PolymorphicInlineSupportMixin, ModelAdmin):
     inlines = (Inline,)
+    extra = 1
