@@ -2,9 +2,10 @@ import pytest
 import re
 import uuid
 
+from django.contrib.auth import get_user_model
 from django.contrib.contenttypes.models import ContentType
 from django.db import models, connection
-from django.db.models import Case, Count, FilteredRelation, Q, Sum, When, F
+from django.db.models import Case, Count, FilteredRelation, Q, Sum, When, Exists, OuterRef
 from django.db.utils import IntegrityError, NotSupportedError
 from django.test import TransactionTestCase
 from django.test.utils import CaptureQueriesContext
@@ -25,6 +26,9 @@ from polymorphic.tests.models import (
     CustomPkInherit,
     Enhance_Base,
     Enhance_Inherit,
+    InlineParent,
+    InlineModelA,
+    InlineModelB,
     InitTestModelSubclass,
     Model2A,
     Model2B,
@@ -82,7 +86,7 @@ from polymorphic.tests.models import (
     SubclassSelectorAbstractConcreteModel,
     SubclassSelectorProxyBaseModel,
     SubclassSelectorProxyConcreteModel,
-    TestParentLinkAndRelatedName,
+    ParentLinkAndRelatedName,
     UUIDArtProject,
     UUIDPlainA,
     UUIDPlainB,
@@ -91,6 +95,10 @@ from polymorphic.tests.models import (
     UUIDResearchProject,
     Duck,
     PurpleHeadDuck,
+    Account,
+    SpecialAccount1,
+    SpecialAccount1_1,
+    SpecialAccount2,
 )
 
 
@@ -450,8 +458,7 @@ class PolymorphicTests(TransactionTestCase):
     def test_onetoone_field(self):
         self.create_model2abcd()
 
-        # FIXME: We should not use base_objects here.
-        a = Model2A.base_objects.get(field1="C1")
+        a = Model2A.objects.non_polymorphic().get(field1="C1")
         b = One2OneRelatingModelDerived.objects.create(one2one=a, field1="f1", field2="f2")
 
         # FIXME: this result is basically wrong, probably due to Django cacheing (we used base_objects), but should not be a problem
@@ -693,7 +700,7 @@ class PolymorphicTests(TransactionTestCase):
         oa.m2m.add(oa)
         oa.m2m.add(ob)
 
-        objects = RelationBase.objects.all()
+        objects = RelationBase.objects.order_by("pk").all()
         assert (
             repr(objects[0])
             == '<RelationBase: id 1, field_base (CharField) "base", fk (ForeignKey) None, m2m (ManyToManyField) 0>'
@@ -952,17 +959,17 @@ class PolymorphicTests(TransactionTestCase):
         assert o.bar == "XYZ"
 
     def test_parent_link_and_related_name(self):
-        t = TestParentLinkAndRelatedName(field1="TestParentLinkAndRelatedName")
+        t = ParentLinkAndRelatedName(field1="ParentLinkAndRelatedName")
         t.save()
-        p = ModelShow1_plain.objects.get(field1="TestParentLinkAndRelatedName")
+        p = ModelShow1_plain.objects.get(field1="ParentLinkAndRelatedName")
 
         # check that p is equal to the
-        assert isinstance(p, TestParentLinkAndRelatedName)
+        assert isinstance(p, ParentLinkAndRelatedName)
         assert p == t
 
         # check that the accessors to parent and sublass work correctly and return the right object
-        p = ModelShow1_plain.objects.non_polymorphic().get(field1="TestParentLinkAndRelatedName")
-        # p should be Plain1 and t TestParentLinkAndRelatedName, so not equal
+        p = ModelShow1_plain.objects.non_polymorphic().get(field1="ParentLinkAndRelatedName")
+        # p should be Plain1 and t ParentLinkAndRelatedName, so not equal
         assert p != t
         assert p == t.superclass
         assert p.related_name_subclass == t
@@ -970,6 +977,29 @@ class PolymorphicTests(TransactionTestCase):
         # test that we can delete the object
         t.delete()
 
+    def test_polymorphic__accessor_caching(self):
+        blog_a = BlogA.objects.create(name="blog")
+
+        blog_base = BlogBase.objects.non_polymorphic().get(id=blog_a.id)
+        blog_a = BlogA.objects.get(id=blog_a.id)
+
+        # test reverse accessor & check that we get back cached object on repeated access
+        self.assertEqual(blog_base.bloga, blog_a)
+        self.assertIs(blog_base.bloga, blog_base.bloga)
+        cached_blog_a = blog_base.bloga
+
+        # test forward accessor & check that we get back cached object on repeated access
+        self.assertEqual(blog_a.blogbase_ptr, blog_base)
+        self.assertIs(blog_a.blogbase_ptr, blog_a.blogbase_ptr)
+        cached_blog_base = blog_a.blogbase_ptr
+
+        # check that refresh_from_db correctly clears cached related objects
+        blog_base.refresh_from_db()
+        blog_a.refresh_from_db()
+
+        self.assertIsNot(cached_blog_a, blog_base.bloga)
+        self.assertIsNot(cached_blog_base, blog_a.blogbase_ptr)
+
     def test_polymorphic__aggregate(self):
         """test ModelX___field syntax on aggregate (should work for annotate either)"""
 
@@ -1422,3 +1452,167 @@ class PolymorphicTests(TransactionTestCase):
         assert pur.color == "blue"
         # issues/615 fixes following line:
         assert pur.home == "Duckburg"
+
+    def test_subqueries(self):
+        PlainA.objects.all().delete()
+        InlineParent.objects.all().delete()
+        InlineModelA.objects.all().delete()
+        InlineModelB.objects.all().delete()
+
+        pa1 = PlainA.objects.create(field1="plain1")
+        PlainA.objects.create(field1="plain2")
+
+        ip1 = InlineParent.objects.create(title="parent1")
+        ip2 = InlineParent.objects.create(title="parent2")
+
+        ima1 = InlineModelA.objects.create(parent=ip1, field1="ima1")
+        ima2 = InlineModelA.objects.create(parent=ip2, field1="ima2")
+        imb1 = InlineModelB.objects.create(parent=ip1, field1="imab1", field2="imb1", plain_a=pa1)
+        imb2 = InlineModelB.objects.create(parent=ip2, field1="imab2", field2="imb2")
+
+        results = InlineModelA.objects.filter(
+            Exists(PlainA.objects.filter(inline_bs=OuterRef("pk")))
+            | Exists(InlineParent.objects.filter(inline_children=OuterRef("pk")))
+        )
+
+        assert ima1 in results
+        assert ima2 in results
+        assert imb1 in results
+        assert imb2 in results
+
+        results = InlineModelA.objects.filter(
+            Exists(PlainA.objects.filter(inline_bs=OuterRef("pk"), field1="plain1"))
+            | Exists(InlineParent.objects.filter(inline_children=OuterRef("pk"), title="parent2"))
+        )
+
+        assert ima1 not in results
+        assert ima2 in results
+        assert imb1 in results
+        assert imb2 in results
+
+        results = InlineModelA.objects.filter(
+            Exists(PlainA.objects.filter(inline_bs=OuterRef("pk")))
+        )
+
+        assert ima1 not in results
+        assert ima2 not in results
+        assert imb1 in results
+        assert imb2 not in results
+
+        results = InlineModelA.objects.filter(
+            Exists(PlainA.objects.filter(inline_bs=OuterRef("pk"))), field1="imab1"
+        )
+
+        assert ima1 not in results
+        assert ima2 not in results
+        assert imb1 in results
+        assert imb2 not in results
+
+        results = InlineModelA.objects.filter(
+            Exists(PlainA.objects.filter(inline_bs=OuterRef("pk"))), InlineModelB___field2="imb2"
+        )
+
+        assert not results
+
+        results = InlineModelA.objects.filter(
+            ~Exists(PlainA.objects.filter(inline_bs=OuterRef("pk"))), InlineModelB___field2="imb2"
+        )
+
+        assert len(results) == 1
+        assert imb2 in results
+
+        PlainA.objects.all().delete()
+        InlineParent.objects.all().delete()
+        InlineModelA.objects.all().delete()
+        InlineModelB.objects.all().delete()
+
+    def test_one_to_one_primary_key(self):
+        # check pk name resolution
+        for mdl in [Account, SpecialAccount1, SpecialAccount1_1, SpecialAccount2]:
+            assert mdl.polymorphic_primary_key_name == mdl._meta.pk.name
+
+        user1 = get_user_model().objects.create(
+            username="user1", email="user1@example.com", password="password"
+        )
+        user2 = get_user_model().objects.create(
+            username="user2", email="user2@example.com", password="password"
+        )
+        user3 = get_user_model().objects.create(
+            username="user3", email="user3@example.com", password="password"
+        )
+        user4 = get_user_model().objects.create(
+            username="user4", email="user4@example.com", password="password"
+        )
+
+        user1_profile = SpecialAccount1_1.objects.create(user=user1, extra1=5, extra2=6)
+
+        user2_profile = SpecialAccount1.objects.create(user=user2, extra1=5)
+
+        user3_profile = SpecialAccount2.objects.create(user=user3, extra1="test")
+
+        user4_profile = SpecialAccount1_1.objects.create(user=user4, extra1=7, extra2=8)
+
+        user1.refresh_from_db()
+        assert user1.account.__class__ is SpecialAccount1_1
+        assert user1.account.extra1 == 5
+        assert user1.account.extra2 == 6
+        assert user1_profile.pk == user1.account.pk
+
+        user2.refresh_from_db()
+        assert user2.account.__class__ is SpecialAccount1
+        assert user2.account.extra1 == 5
+        assert user2_profile.pk == user2.account.pk
+        assert not hasattr(user2.account, "extra2")
+
+        user3.refresh_from_db()
+        assert user3.account.__class__ is SpecialAccount2
+        assert user3.account.extra1 == "test"
+        assert user3_profile.pk == user3.account.pk
+        assert not hasattr(user3.account, "extra2")
+
+        user4.refresh_from_db()
+        assert user4.account.__class__ is SpecialAccount1_1
+        assert user4.account.extra1 == 7
+        assert user4.account.extra2 == 8
+        assert user4_profile.pk == user4.account.pk
+
+        assert get_user_model().objects.filter(pk=user2.pk).delete() == (
+            3,
+            {"tests.SpecialAccount1": 1, "tests.Account": 1, "auth.User": 1},
+        )
+
+        assert SpecialAccount1.objects.count() == 2
+        assert Account.objects.count() == 3
+
+        remaining = get_user_model().objects.filter(
+            pk__in=[user1.pk, user2.pk, user3.pk, user4.pk]
+        )
+        assert remaining.count() == 3
+        for usr, expected in zip(
+            remaining.order_by("pk"), (user1_profile, user3_profile, user4_profile)
+        ):
+            assert usr.account == expected
+
+        assert get_user_model().objects.filter(pk__in=[user3.pk]).delete() == (
+            3,
+            {"tests.SpecialAccount2": 1, "tests.Account": 1, "auth.User": 1},
+        )
+
+        assert Account.objects.count() == 2
+
+        assert SpecialAccount1_1.objects.all().delete() == (
+            6,
+            {"tests.SpecialAccount1_1": 2, "tests.SpecialAccount1": 2, "tests.Account": 2},
+        )
+
+        assert Account.objects.count() == 0
+
+        remaining = get_user_model().objects.filter(pk__gte=user1.pk)
+        assert remaining.count() == 2
+        for usr in remaining:
+            assert not hasattr(usr, "account")
+
+        assert get_user_model().objects.filter(pk__in=[user1.pk, user4.pk]).delete() == (
+            2,
+            {"auth.User": 2},
+        )
